if Array.length Sys.argv < 2 then
  raise (Failure ("Missing parameters. Call with: " ^ Sys.argv.(0) ^ " <spec file> [<C files>]\nIf no C file is supplied, all .c files in the cwd will be checked."));

let specFile = Sys.argv.(1) in
print_string "Parsing sources....\n";
flush stdout;
let fnames =
  if Array.length Sys.argv >= 3 then
    Array.to_list (Array.sub Sys.argv 2 (Array.length Sys.argv - 2))
  else
    (let dir = Unix.opendir "." in
    let rec getFilenames () =
      try
	let fname = Unix.readdir dir in
	if (String.length fname >= 2 && compare ".c" (String.sub fname (String.length fname - 2) 2) == 0)
	    && compare "merged.c" fname != 0
	    && compare "instrumented.c" fname != 0 then
	  fname :: getFilenames ()
	else
	  getFilenames ()
      with End_of_file -> [] in
    let fnames = getFilenames () in
    Unix.closedir dir;
    fnames) in
let files = List.map Common.parseFile fnames in
let file = Mergecil.merge files "merged.c" in

let merged = open_out "merged.c" in
Common.outputFile merged file;
close_out merged;

(*print_string "Initial code:\n";
Common.printFile file;*)

Rmtmps.removeUnusedTemps file;

let spec = Spec.parseSpec specFile in

print_string "Spec read\n";
flush stdout;

(*print_string "\nSpec:\n";
Spec.printSpec spec;*)

let file = Spec.instrument spec file in

(*print_string "\Before rmtmps:\n";
Common.printFile file;

Rmtmps.removeUnusedTemps file;*)

(*print_string "\nInstrumented code:\n";
Common.printFile file;*)

let outf = open_out "instrumented.c" in
Common.outputFile outf file;
close_out outf;

print_string "Done\n"
