(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: misc.mli,v 1.4 2005/06/20 22:59:49 jhala Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module provides some miscellaneous useful helper functions.
 *)

(*module type INTSET = 
  sig
    include Set.S
    type elt = int
  end*)
  
(** swap a pair *)
val swap : ('a * 'b) -> ('b * 'a)

(* playing with triples *)
(** get the first element of a triple *)
val fst3 : ('a * 'b * 'c) -> 'a

(** get the second element of a triple *)
val snd3 : ('a * 'b * 'c) -> 'b

(** get the third element of a triple *)
val thd3 : ('a * 'b * 'c) -> 'c

  
(** adds an element to a list representing a set *)
val add : 'a list -> 'a -> 'a list

(** union two lists representing sets *)
val union : 'a list -> 'a list -> 'a list

(** intersect two lists representing sets *)
val intersection : 'a list -> 'a list -> 'a list

val sort_and_intersect : 'a list -> 'a list -> 'a list
(** removes the duplicate elements of a list *)
val compact : 'a list -> 'a list

  
(** delete an item from a list *)
val list_remove : 'a list -> 'a -> 'a list

(** return the Single item in the list satisfying f -- else throw exn *)
val list_filter_unique : ('a -> bool) -> 'a list -> 'a

(** set difference using lists to represent sets *)
val difference : 'a list -> 'a list -> 'a list
val proj_difference_l : ('a * 'b) list -> 'a list -> ('a * 'b) list
val proj_difference_r : 'a list -> ('a * 'b) list -> 'a list


(** set containment using lists *)
val subsetOf : 'a list -> 'a list -> bool

(** sort, and then compact *) 
val sort_and_compact : 'a list -> 'a list
  
(** set containment of sorted lists *)  
val sorted_subsetOf : 'a list -> 'a list -> bool
  
(** cartesian product *)
val cartesianProduct : 'a list list -> 'a list list

(** given a list of strings, produce a single string that is the
   comma-separated list of the original strings *)
val strList : string list -> string

(** returns the to_string function corresponding to the given
   pretty-printer *)
val to_string_from_printer :
  (Format.formatter -> 'a -> unit) -> 'a -> string

(** given a printer for a type t, returns a printer for lists of elements of
   type t *)
val list_printer_from_printer :
  (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a list -> unit

(** returns true if the char is a digit (0-9) *)
val is_digit : char -> bool

(** a fault tolerant map for lists.
    map_exn takes an exception {i ex} to be caught, a function {i f: 'a -> 'b)} 
    and an {i 'a list}, and produces a {i 'b list} by
    applying function {i f} to each element of the list.
    If evaluating the function raises exception {i ex},
    the corresponding element is omitted in the result.
*)
val map_exn : exn -> ('a -> 'b) -> 'a list -> 'b list


(** the civilized way to do the above. apparently *)
(** map_partial f l *)
val map_partial : ('a -> 'b option) -> 'a list -> 'b list

(** given a predicate {i P} and a list {i l}, return two
   sublists of {i l}: those that satisfy {i P}, and those
   that do not.
*)
val filter_cut : ('a -> bool) -> 'a list -> ('a list * 'a list)

(** returns the truncated list containing the first n elements Raise Failure "nth" if list is too short *)
val truncateList : 'a list -> int -> 'a list

(** is_prefix pref str checks if pref is a prefix of string str *)
val is_prefix : string -> string -> bool
(** is_suffix suffix str checks if suffix is a suffix of string str *)
val is_suffix : string -> string -> bool

(** bounded_chop s chopper i breaks s into at most i pieces ... *)
val bounded_chop : string -> string -> int -> string list

(** is_substring s subs is true iff subs is a substring of s *)
val is_substring : string -> string -> bool

(** takes a list of strings and subs each occurrence to get the result *)
val substitute_substrings : (string * string) list -> string -> string

(** k_suffix i (s0.s1.s2...sn) returns s_{i+1} ...s_n *) 
val k_suffix : int -> string -> string
(** given a prefix and a string, return the suffix of
    the string after the prefix, i.e., if s = p.s', then
    {i chop_after_prefix p s} is {i s'} *)
val chop_after_prefix : string -> string -> string 
(** given a suffix and a string, return the prefix of
    the string before the suffix, i.e., if s = s'.p, then
    {i chop_before_suffix p s} is {i s'} *)
val chop_before_suffix : string -> string -> string 

val unzip : ('a * 'b) list -> ('a list * 'b list)


val hashtbl_map_partial : ('a -> 'b -> 'c option) -> ('a,'b) Hashtbl.t -> 'c list

(** f: (a->b->bool) t = ('a,'b) Hashtbl.t, what else can I say ? *)
val hashtbl_exists : ('a -> 'b -> bool) -> ('a, 'b) Hashtbl.t -> bool

val hashtbl_exists_key : ('a -> bool) -> ('a,'b) Hashtbl.t -> bool

(** returns the domain of a hashtbl *)
val hashtbl_keys : ('a, 'b) Hashtbl.t -> 'a list

(** returns the range of a hashtbl *)
val hashtbl_data : ('a, 'b) Hashtbl.t -> 'b list

(** hashtbl_filter f tbl returns the list a, s.t. f a b = true, where a -> b in tbl *)
val hashtbl_filter : ('a -> 'b -> bool) -> ('a, 'b) Hashtbl.t -> 'a list
 
(** hashtbl_filter_tbl f tbl removes all key,data
 *  bindings from tbl where (f key data = false) *) 
val hashtbl_filter_tbl : ('a -> 'b -> bool) -> ('a, 'b) Hashtbl.t -> unit


(** returns the pairs of a hashtbl *)
val hashtbl_to_list : ('a, 'b) Hashtbl.t -> ('a * 'b) list

(** returns a hashtbl whose keys are the elements of the list *)
val hashtbl_of_list : int -> ('a list) -> ('a, bool) Hashtbl.t

val hashtbl_update : ('a, 'b list) Hashtbl.t -> 'a -> 'b -> unit

val hashtbl_check_update : ('a, 'b list) Hashtbl.t -> 'a -> 'b -> bool

val hashtbl_fun_check_update : ('b -> 'b -> bool) -> ('a, 'b list) Hashtbl.t -> 'a -> 'b -> bool

  
val hashtbl_delete : ('a, 'b list) Hashtbl.t -> 'a -> 'b -> unit

(** Hashtbl union: {[hashtbl_addtable t1 t2]} adds all data of {i t2} to {i t1} *)
val hashtbl_addtable : ('a, 'b) Hashtbl.t -> ('a, 'b) Hashtbl.t -> unit 

(* val search_list : ('a -> 'b option) -> 'a list -> b' option *)


(** given a predicate and a list, break the list at the first point an
    element satisfies the predicate. The element satisfying the predicate
    is in the second part.
*)
val chop_list : ('a -> bool) -> ('a list) -> ('a list * 'a list)
 
(* chop s chopper returns Some([x;y;z...]) if s = x.chopper.y.chopper ..., None o.w. *)
val chop : string -> string -> string list

(** Given a list of 'a option, get all the elements of the form Some v *)
val list_gather_option : ('a option list) -> ('a list)

(** Add an elemt to a list, if the element is not actually there *)
val cons_if : 'a -> 'a list -> 'a list
  
(** find the sum of a list of integers *)
val list_add : int list -> int

(** unpackage a string option value *)
val string_option_to_string : string option -> string

(** turn a list of strings into a simplify predicate *)
(* simplify_boolean_unroll list op base_case *)
val simplify_boolean_unroll : string list -> string -> string -> string 

(* only for nonempty lists *)
val list_max : int list -> int

(* the name says it all ... write_to_file fname string-to-write *)
val write_to_file : string -> string -> unit

(* write to file with a list of strings *)
val write_list_to_file : string -> string list -> unit

(* the name says it all ... append_to_file fname string to write *)
val append_to_file : string -> string -> unit

(* returns the list with the last element removed *)
val chop_last : 'a list -> 'a list

(** makes all possible pairs out of a list of elements *)
val make_all_pairs : 'a list -> ('a * 'a) list

(** concatenate the elts of list using sep as separator: string_list_cat sep slist *)
val string_list_cat : string -> string list -> string 

(** gives the cross_product of the two lists *)
val cross_product : 'a list -> 'b list -> ('a * 'b) list

(** resets all entries of the given array to the given value *)
val array_reset : 'a array -> 'a -> unit

(** string_list_of_file fname returns the file as a list of strings (delim by \n ) *)
val string_list_of_file : string -> string list

(** takes a string and returns the list of words that appear in it *)
val words_of_string : string -> string list

(** grep_file sfile grepfile: returns the "lines" of sfile that contain a word in grepfile *)
val grep_file : string -> string -> string list list

(** run_on_table f t : produces a list which is the result of doing a map using f on the hashtbl t *)
val run_on_table : ('a -> 'b -> 'c) -> ('a, 'b) Hashtbl.t -> 'c list

(** list_last l : returns the last element of the list l and fails if the list is empty *)
val list_last : 'a list -> 'a

(** bool_to_string b : returns a string from the bool *)
val bool_to_string : bool -> string

(** list_cut i l1 : returns a pair of 'a, 'a list where a' is the i+1 th elt from the 
start of the list and the 'a list is the 1 through i'th elts in reverse *)
val list_cut : int -> 'a list  -> ('a * ('a list))

(** Like list_cut, but the new first return tuple component is all elements coming after a'
 *  in the list, in their original order. *)
val list_cut' : int -> 'a list  -> (('a list) * 'a * ('a list))

(** min l returns the smallest elt in l and -1 if l is empty, assumes that all elts are >= 0 *)
val min : int list -> int

(** unzip_joined [e1;e2;e3 ... en] returns ([e1,e3,e5,...],[e2,e4,e6...]) *)
val unzip_joined : 'a list -> ('a list * 'a list)

(** repeat_fn f i just calls f, i times *)
val repeat_fn : (unit -> 'a) -> int -> unit

(** powerSetb l : compute the power set of the list l *)
val powerSet : 'a list -> ( (bool * 'a) list ) list

(** power a b: exponentiation -- compute a^b *)
val power : int -> int -> int

(** map the elements in order (the function is assumed to be stateful) *)
val list_ordered_map : ('a -> 'b) -> 'a list -> 'b list

val maximal : ('a -> 'a -> bool) -> 'a list -> 'a list

val paren_match : ('a -> string) -> 'a array -> ((int -> int) * (int -> int))

val list_tabulate : (int -> 'a) -> int -> int -> 'a list

val kcombine : 'a list list -> 'a list

val partial_combine : 'a list -> 'b list -> ('a * 'b) list
  
val make_list : int -> int list

val hashtbl_forall : ('a -> 'b -> bool) -> ('a, 'b) Hashtbl.t -> bool

val cut_list_at_k : 'a list -> int -> ('a list * 'a list)

val get_first_array : 'a array -> ('a -> bool) -> int -> int option

val get_first : ('a -> 'b option) -> 'a list -> 'b option

val get_first_cont_suffix : ('a -> bool) -> 'a list -> 'a list

val array_of_list2 : 'a list list -> 'a array array

val array_filter : ('a -> bool) -> 'a array -> int list

val array_filteri : (int -> bool) -> 'a array -> 'a list

val array_select2 : 'a array -> int list list -> 'a list list

val make_list_pairs : 'a list -> ('a * 'a) list 

val hd_ll : 'a list list -> 'a option

val string_of_int_list : int list -> string

val boolean_list_leq : bool list -> bool list -> bool

(** get_binary k n gets the binary representation of k to max(log k, n) bits *)
val get_binary : int -> int -> int list

(** ignore_list_iter f l = ignore (List.map f l) -- cant say it more succintly than that! *)
val ignore_list_iter : ('a -> 'b) -> 'a list -> unit

(**  map_transpose f l : 
- map_transpose List.length [[1;2;3;4];[5;6;7;8];[9;9;9;9]] ;;
- : int list = [3; 3; 3; 3]
*)

val map_transpose : ('a list -> 'b) -> 'a list list -> 'b list

val queue_to_list : 'a Queue.t -> 'a list

val option_to_list : 'a option -> 'a list

(**   list_mask f l = [e_i1;e_i2;...] where f.ij is true *)
val list_mask : (int -> bool) -> 'a list -> 'a list

(** array_counti f a returns | { (i,a[i]) | f.i.a[i] is true } | *)  
val array_counti : (int -> 'a -> bool) -> 'a array -> int  

(* nonempty_intersect compare l1 l2 tells you if there is a non-zero intersection .. *)  
val nonempty_intersect : ('a -> 'a -> int) -> 'a list -> 'a list -> bool  

(* a generalization of the above -- in addition require the intersectee to satisfy f *)
val nontrivial_intersect : ('a -> bool) -> ('a -> 'a -> int) -> 'a list -> 'a list -> bool 

(* exists_file fname : is true iff there is a file called fname in the current directory *)
val exists_file : string -> bool

(* shady hack -- would have used type constructors but circular dependency etc. *)  
val allLvals_t_c : int
val scopeLvals_t_c : int
val traceLvals_t_c : int
  
val strIntList : int list -> string

(** takes: next_fn, seed (start list), returns fp, together with "shortest path
lengths" *)
val compute_fixpoint : ('a -> 'a list) -> 'a list -> ('a * int) list

(** like the above, but takes a depth bound as well *)
val compute_fixpoint_bounded : int -> ('a -> 'a list) -> 'a list -> ('a * int) list

(* accumulate f binop seed [a1;a2 ...] .... returns (seed binop (f a1) binop (f a2) ...) *)
val accumulate : ('a -> 'b) -> ('c -> 'b -> 'c) -> 'c -> 'a list -> 'c

val list_n : 'a list -> int -> 'a list

val array_to_list_n : 'a array -> int -> 'a list

val toString_list : ('a -> string) -> 'a list -> string

val list_min_combine : ('a list) -> ('b list) -> ('a * 'b) list

val list_argmin : ('a -> int) -> 'a list -> ('a * int)

module IntSet : Set.S with type elt = int

val hashints_check_union : ('a, IntSet.t) Hashtbl.t -> 'a -> IntSet.t -> bool

val list_exists_cross2 : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool

val group_pairwise : 'a list -> ('a * 'a) list 

(** base_convert b n : is the equivalent of n in base 10 in base b *)
val base_convert : int -> int -> int list


(** converts 0 --> a, 1 --> b, ...,26 --> ba etc. *)
val ascii_string_of_int : int -> string
