(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
{
module E = VampyreErrormsg
module I = Input
open Lfparse

let lexerror msg lexbuf = 
  E.s (E.unimp "%s" msg)

}

let digit    = ['0'-'9']
let letdig   = ['0'-'9' 'a'-'z' 'A'-'Z' '_' '-' '$' '#' '@' '!' '+' '=' 
                '<' '>' ',' '?' '\'']

let capital  = ['A'-'Z']
let small    = ['a'-'z' '$' '_']
let ws       = [' ' '\009' '\012']
let pathname = ['a'-'z' 'A'-'Z' '0'-'9' '.' '/' '\\' '-']

rule token = parse
    '\r'                { token lexbuf}
  | '\n'		{ begin
			    E.startNewline (Lexing.lexeme_end lexbuf);
			    token lexbuf 
			  end }
  | "%!Decl"		{ BEGINDECL }
  | "%!EndDecl"		{ ENDDECL }
  | "%!COMM"[^'!''\n']*'\n'
      			{ begin
			    E.startNewline (Lexing.lexeme_end lexbuf);
			    let comm = Lexing.lexeme lexbuf in
			      Comment (String.sub comm 0 
                                         ((String.length comm) - 1))
			  end }
  | '%'[^'\n']*'\n'	{ begin
			    E.startNewline (Lexing.lexeme_end lexbuf);
			    token lexbuf 
			  end }
  | "IMPORT"[^'\n']*'\n' { begin
                            E.startNewline (Lexing.lexeme_end lexbuf);
                            token lexbuf
                          end}
  | ws			{ token lexbuf }
  | '('			{ LPAREN }
  | ')'			{ RPAREN }
  | '['			{ LBRACKET }
  | ']'			{ RBRACKET }
  | '{'			{ LBRACE }
  | '}'			{ RBRACE }
  | "->"		{ ARROW }
  | '.'			{ DOT }
  | ':'			{ COLON }
  | '+'			{ PLUS }
  | '-'			{ MINUS }
  | '*'			{ TIMES }
  | "<<"		{ Id "<<" }
  | ">>"		{ Id ">>" }
  | "<"			{ LT }
  | ">"			{ GT }
  | "<="		{ LEQ }
  | ">="		{ GEQ }
  | '='			{ EQ }
  | "<>"		{ NEQ }
  | "mod"		{ MOD }
  | "div"		{ DIV }
  | "floor"		{ FLOOR }
  | "ceil"		{ CEIL }
  | "true"		{ TRUE }
  | "false"		{ FALSE }
  | "/\\"		{ AND }
  | "\\/"		{ OR }
  | "~"                 { NOT }
  | "=>"		{ IMPL }
  | "all"		{ ALL }
  | "exp"		{ EXP }
  | "pred"		{ PRED }
  | "pf"		{ PF }
  | '~'?(digit)+	{ let str = Lexing.lexeme lexbuf in
			  let len = String.length str in
			  let zero = Char.code '0' in
			  let rec accum a d =
			    let acc c = a + (d * ((Char.code c) - zero)) in
			    function
			      0 -> let c = str.[0] in
				   if c='~' then - a else (acc c)
			    | i -> accum (acc str.[i]) (d * 10) (i - 1)
			  in
			  Num (accum 0 1 (len-1)) }
  | (small)letdig*	{ Id (Lexing.lexeme lexbuf) }
  | (capital)letdig*	{ Var (Lexing.lexeme lexbuf) }
  | eof			{ EOF }
  | _			{ 
                          begin
                            lexerror ("Illegal Character '" ^ 
                                      (Lexing.lexeme lexbuf)) lexbuf;
			    token lexbuf
			  end }

